#region license
// Copyright (C) 2010-2015 Mikael Lyngvig (mikael@lyngvig.org).  All Rights Reserved.
//
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
// conditions are met:
//
//     * Redistributions of source code must retain the above copyright notice, this list of conditions and the disclaimer below.
//     * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
//       disclaimer in the documentation and/or other materials provided with the distribution.
//     * Neither the name of Mikael Lyngvig nor the names of its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
// BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
// SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#endregion

/** \file
 *  The definition of a reference to a non-terminal AST node (typically written <tt>... Name ...</tt>).
 */

using System.Collections.Generic;

using Org.Braceless.Bracell.Toolbox;

namespace Org.Braceless.Bracell.Syntax
{
    /** Represents a reference to a non-terminal on the \b left side of the equal sign: <tt>Node = ... Name ...</tt>. */
	public class ReferenceNode : Node
	{
		private string _name;
		public override string Name
		{
			get { return _name; }
		}

		public ReferenceNode(Cursor cursor, string name):
			base(NodeKind.Reference, cursor)
		{
			_name = name;
		}

		/** Computes the \c TAIL(x) function of the specified node, which is a node in the below list of a \c ReferenceNode. */
		public Node[] ComputeTail(uint index)
		{
			// find this node in the parent's list of children
			Node[] siblings = this.Below;
			uint i;
			for (i = index; i < siblings.Length; i++)
			{
				if (siblings[i] == this)
				{
					i += 1;
					break;
				}
			}
			if (i >= siblings.Length)
				return new Node[0];

			// create array of the actual tail of 'this' node
			Node[] result = new Node[siblings.Length - i + 1];
			for (uint j = i; j < siblings.Length; j++)
				result[j - i] = siblings[j];
			if (result.Length == 0)
				throw new System.Exception("Internal error: Null length tail");
			return result;
		}

		/** Inspired by Crafting a Compiler, Global Edition, page. 188 (figure 5.13).
		 *  Albeit rewritten to perform the analysis AFTER the transformation to Standard Grammar Language.
		 */
		private uint LongestCommonPrefix(Node that_node)
		{
			if (that_node == null)
				return 0;

			ReferenceNode that = that_node as ReferenceNode;
			if (that == null)
				return 0;

			// find the
			uint i;
			for (i = 0; i < System.Math.Min(this.Below.Length, that.Below.Length); i++)
			{
				if (!this.Below[i].Equals(that.Below[i]))
					break;
			}
			return i;
		}

		public override bool Equals(object that_object)
		{
			if (!base.Equals(that_object))
				return false;

			ReferenceNode that = that_object as ReferenceNode;
			return this.Name == that.Name;
		}

        public override int GetHashCode()
        {
            int result = base.GetHashCode();
            result ^= _name.GetHashCode();
            return result;
        }

		public override void Visit(Visitor that)
		{
			that.Visit(this);
		}
	}
}
